# Copyright 2021 Nokia

import contextlib
import copy
import json

from enum import Enum, auto

from lxml import etree

from .errors import *
from .identifier import Identifier
from .model import Model
from .wrappers import *
from .yang_type import *

class Token:
    def __init__(self):
        self.kind: Optional["ModelWalker._TokenKind"] = None
        self.value = ''

class ModelWalker:
    """API representation for walking model tree.

    .. Reviewed by TechComms 20210712
    """
    _expected_dds = (Model.StatementType.container_, Model.StatementType.list_, Model.StatementType.leaf_, Model.StatementType.leaf_list_, Model.StatementType.augment_, Model.StatementType.notification_, Model.StatementType.rpc_, Model.StatementType.input_, Model.StatementType.output_, Model.StatementType.choice_, Model.StatementType.case_)
    _recursive_visited_dds = (Model.StatementType.module_, Model.StatementType.submodule_, Model.StatementType.uses_, Model.StatementType.grouping_, Model.StatementType.augment_)

    def __init__(self, model:Model):
        self.path = []
        self.keys = []
        while model.parent is not None:
            if model.data_def_stm in self._expected_dds:
                self.path.insert(0, model)
                self.keys.insert(0, dict())
            model = model.parent
        self.model = model

    @property
    def current(self):
        return self.path[-1] if len(self.path) > 0 else self.model

    @property
    def local_keys(self):
        return self.keys[-1]

    @property
    def is_root(self):
        return not self.path

    @property
    def is_state(self):
        return not self.current.config

    def go_to_parent(self):
        if not self.path:
            raise make_exception(pysros_err_cannot_call_go_to_parent)

        self.path.pop()
        self.keys.pop()

    def go_to_child(self, child_name:Union[str, Identifier]):
        child = self._get_child(child_name)
        self.path.append(child)
        self.keys.append(dict())
        return child

    def go_to(self, path: ModelPath, module: Optional[str] = None):
        for p in path._path:
            if not p.is_lazy_bound():
                try:
                    self.go_to_child(p)
                except SrosMgmtError as e:
                    raise InvalidPathError(*e.args) from None
                continue

            if p.name == '..':
                self.go_to_parent()
                continue

            namespace = self.path[-1].name.prefix if self.path else module
            assert namespace

            try:
                self.go_to_child(
                    Identifier(namespace, p.name))
            except SrosMgmtError as e:
                raise InvalidPathError(*e.args) from None

    def check_child_field_value(self, name:Union[str, Identifier], value):
        with self.visit_child(name):
            return self.check_field_value(value)

    def check_field_value(self, value):
        if self.get_dds() == Model.StatementType.leaf_list_:
            return all(self.get_type().check_field_value(i) for i in value)
        elif self.get_dds() == Model.StatementType.leaf_:
            return self.get_type().check_field_value(value)
        else:
            assert False, "Checking field value for non-field walker"

    def get_parent(self):
        res = self.__class__(self.model)
        if self.path:
            res.path = self.path[:-1]
            res.keys = self.keys[:-1]
        return res

    def get_child(self, child_name:Union[str, Identifier]):
        res = self.copy()
        res.path.append(self._get_child(child_name))
        res.keys.append(dict())
        return res

    def copy(self):
        res = self.__class__(self.model)
        res.path = self.path[:]
        res.keys = copy.deepcopy(self.keys)
        return res

    def go_to_last_with_presence(self):
        while self.path and not self.has_explicit_presence():
            self.go_to_parent()

    def has_explicit_presence(self):
        types_with_presence = (Model.StatementType.leaf_, Model.StatementType.leaf_list_)
        return self.get_dds() in types_with_presence or \
               self.get_dds() == Model.StatementType.list_ and not self.has_missing_keys() or \
               self.get_dds() == Model.StatementType.container_ and self.current.presence_container

    def has_missing_keys(self):
        return self.get_dds() == Model.StatementType.list_ and len(self.keys[-1]) != len(self.current.local_keys)

    def dict_keys(self, value):
        for i in value.keys():
            if not isinstance(i, tuple):
                i = (i,)
            if (len(self.get_local_key_names()) != len(i)):
                return False
            if not all(self.check_child_field_value(key_name, key_val) for key_name, key_val in zip(self.get_local_key_names(), i)):
                return False
        return all(isinstance(v, (dict, Container)) for v in value.values())

    def entry_keys(self, value:"EntryKeysDictProxy"):
        def correct_key(entry:"EntryKeysDictProxy", key):
            if key in entry and self.check_child_field_value(key, entry[key]):
                return True
            return False

        local_keys = [Identifier(self.get_name().prefix, k) for k in self.get_local_key_names()]
        return all(correct_key(value, k) for k in local_keys)

    def as_model_type(self, s):
        assert self.get_dds() in (Model.StatementType.leaf_, Model.StatementType.leaf_list_)
        if isinstance(s, list):
            res = [self.get_type().to_value(p) for p in s]
        else:
            res = [self.get_type().to_value(s)]

        if self.get_dds() == Model.StatementType.leaf_:
            res = res[0]
        return res

    def dump_json(self):
        def _dump_json(self, parent_data):
            if self.get_dds() in self._expected_dds:
                data = {"name" : self.get_name().name,
                        "dds"  : self.get_dds().name}
                if self.get_dds() != Model.StatementType.leaf_:
                    data["children"] = []
                if self.get_dds() == Model.StatementType.leaf_:
                    data["type"] = self.get_type().json_name()
                parent_data["children"].append(data)
                parent_data = data

            children = list(self.current.children)
            while children:
                child = children.pop()
                if child.data_def_stm in (Model.StatementType.module_, Model.StatementType.submodule_):
                    children.extend(child.children)
                    continue
                if child.data_def_stm in (Model.StatementType.container_, Model.StatementType.list_, Model.StatementType.leaf_, Model.StatementType.leaf_list_, Model.StatementType.choice_, Model.StatementType.case_, ):
                    with self.visit_child(child.name.name):
                        _dump_json(self, parent_data)
        root_data = {"children":[]}
        _dump_json(self, root_data)
        return json.dumps(root_data, indent=4)



    def as_child_model_type(self, child, s):
        with self.visit_child(child):
            return self.as_model_type(s)

    @contextlib.contextmanager
    def visit_child(self, child:Union[str, Identifier]):
        self.go_to_child(child)
        try:
            yield
        finally:
            self.go_to_parent()

    def get_child_type(self, child_name:Union[str, Identifier]):
        return self._get_child(child_name).yang_type

    def get_child_dds(self, child_name:Union[str, Identifier]):
        return self._get_child(child_name).data_def_stm

    def get_child_name(self, child_name:Union[str, Identifier]):
        return self._get_child(child_name).name

    def get_type(self):
        return self.current.yang_type

    def get_dds(self):
        return self.current.data_def_stm

    def get_name(self):
        return self.current.name

    def get_local_key_names(self):
        return self.current.local_keys

    def has_local_keys_specified(self):
        return len(self.get_local_key_names()) == len(self.local_keys)

    def has_local_key_named(self, name):
        return name in self.get_local_key_names()

    def has_field_named(self, name):
        return self.has_child(name) and self.get_child_dds(name) in (Model.StatementType.leaf_, Model.StatementType.leaf_list_)

    @property
    def is_local_key(self):
        return self.get_name().name in self.get_parent().get_local_key_names()

    def has_child(self, child_name:Union[str, Identifier] = None):
        if child_name:
            try:
                self._get_child(child_name)
            except:
                return False
            else:
                return True
        else:
            return self.current.has_children

    class _TokenKind(Enum):
        string  = auto()
        symbol  = auto()

    @classmethod
    def _tokenize(cls, string):
        for i in cls._tokenize_(string):
            yield i

    @classmethod
    def _tokenize_(cls, string):
        res = []
        iterator = iter(string)
        while True:
            try:
                i = next(iterator)
            except StopIteration:
                if res:
                    yield (cls._TokenKind.string, "".join(res))
                return
            if i in "[]=/":
                if res:
                    yield (cls._TokenKind.string, "".join(res))
                    res = []
                yield (cls._TokenKind.symbol, i)
                continue
            if res:
                res.append(i)
                continue
            if i == '"':
                while True:
                    try:
                        i = next(iterator)
                        if i == '"':
                            break
                        res.append(i)
                    except StopIteration:
                        raise make_exception(pysros_err_unended_quoted_string)
                yield (cls._TokenKind.string, "".join(res))
                res = []
                continue

            if i == "'":
                while True:
                    try:
                        i = next(iterator)
                        if i == "'":
                            break
                        res.append(i)
                    except StopIteration:
                        raise make_exception(pysros_err_unended_quoted_string)
                yield (cls._TokenKind.string, "".join(res))
                res = []
                continue
            res.append(i)

    @classmethod
    def user_path_parse(cls, *args, **kwargs):
        res = cls.path_parse(*args, **kwargs)
        assert isinstance(res, ModelWalker)
        if res.is_root:
            raise make_exception(pysros_err_root_path)
        for elem in res.path:
            if elem.is_region_blocked:
                raise make_exception(pysros_err_unknown_element, element=elem.name.name)
        return res

    @classmethod
    def path_parse(cls, model_root, path_string):
        if not isinstance(path_string, str):
            raise make_exception(pysros_err_path_should_be_string)

        if path_string == "/":
            return cls(model_root)

        if not path_string.startswith('/'):
            if not path_string:
                raise make_exception(pysros_err_empty_path)
            raise make_exception(pysros_err_not_found_slash_before_name)
        if path_string.endswith('/'):
            raise make_exception(pysros_err_invalid_identifier)

        iterator = iter(cls._tokenize(path_string))
        def next_token(*accepts, err,  **kwarg):
            """Verify a next token from path and return tuple(kind, token) if no errors.
                Otherwise, raise make_exception(err, **kwarg).

:param accepts: Contains a list of the following:
    - one of '/', '[', ']' or '=' - Function accepts specified symbol.
    - 'string' - Function accepts a string, not a symbol.
    - None - Function accepts end of string.

            .. Reviewed by TechComms 20210713
            """
            n = next(iterator, None)

            if n is None:
                if None in accepts:
                    return None, None
            elif n[0] == cls._TokenKind.symbol:
                if n[1] in accepts:
                    return n
            elif 'string' in accepts:
                return n
            raise make_exception(err, **kwarg)

        res = cls(model_root)
        is_root_element = True
        missing_keys = set()
        elem = 'root'

        while True:
            _, i = next_token('/', '[', None, err = pysros_err_not_found_slash_before_name)
            if i == None:
                if missing_keys and len(missing_keys) != len(res.current.local_keys):
                    raise make_exception(pysros_err_missing_keys, element=elem)
                break
            if i == "/":
                if missing_keys:
                    raise make_exception(pysros_err_missing_keys, element=elem)
                _, elem = next_token('string', None, err = pysros_err_invalid_identifier)
                try:
                    res.go_to_child(elem)
                    missing_keys.update(res.current.local_keys)
                    is_root_element = False
                except Exception as e:
                    raise make_exception(pysros_err_unknown_element, element=elem) from None
                continue
            if i == "[":
                _, key_name = next_token('string', err = pysros_err_invalid_identifier)
                next_token('=', err = pysros_err_expected_equal_operator)
                _, value = next_token('string', err = pysros_err_invalid_identifier)
                next_token(']', err = pysros_err_expected_end_bracket)
                if key_name not in missing_keys:
                    try:
                        res._get_child(key_name)
                    except Exception as _:
                        raise make_exception(pysros_err_unknown_key, key_name = key_name)
                    raise make_exception(pysros_err_cannot_specify_non_key_leaf)
                missing_keys.remove(key_name)
                res.local_keys[key_name] = value
        return res

    @classmethod
    def tokenize_path(cls, path: str, absolute_schema_id: bool):
        # absolute schema id has form: absolute-schema-nodeid = 1*("/" node-identifier)
        # leaf-ref form allows also predicates and a ".." inside paths
        if not path.startswith(('/', '../')):
            raise make_exception(pysros_err_invalid_yang_path, path=path)

        _end = None, ''
        t = Token()

        iterator = iter(cls._tokenize_(path))
        def next_token(t):
            t.kind, t.value = next(iterator, _end)
        next_token(t)
        if t.value == '/':
            # skip '/' in absolute path
            next_token(t)

        while True :
            if t.kind != cls._TokenKind.string:
                raise make_exception(pysros_err_invalid_yang_path, path=path)
            yield t.value
            next_token(t)

            while t.value == '[' and not absolute_schema_id:
                while t.value not in ('', ']'):
                    next_token(t)
                    if t.kind is None:
                        return
                next_token(t)

            if t.value == '/':
                next_token(t)
                continue

            if t.kind is None:
                return

            raise make_exception(pysros_err_invalid_yang_path, path=path)

    def __str__(self):
        return "/" + "/".join(str(name.name) + "".join(f"[{key}={value}]" for key, value in keys.items()) for name, keys in zip(self.path, self.keys))

    def _get_child(self, child_name:Union[str, Identifier]):
        children = list(self.current.children)
        while children:
            child = children.pop()
            if child.name == child_name and child.data_def_stm in self._expected_dds and self._is_allowed(child):
                return child
            if child.data_def_stm in self._recursive_visited_dds:
                children.extend(child.children)
        raise make_exception(pysros_err_unknown_child, child_name=child_name, path=self._get_path())

    def recursive_walk(self, *, enter_fnc=None, leave_fnc=None):
        children = list(self.current.children)
        while children:
            child = children.pop()
            if child.data_def_stm in self._expected_dds and self._is_allowed(child):
                self.path.append(child)
                self.keys.append(dict())
                enter_fnc and enter_fnc(self)
                self.recursive_walk(enter_fnc=enter_fnc, leave_fnc=leave_fnc)
                leave_fnc and leave_fnc(self)
                self.path.pop()
                self.keys.pop()
            if child.data_def_stm in self._recursive_visited_dds:
                children.extend(child.children)

    def _get_path(self):
        return " ".join(str(model.name) for model in self.path)

    def _is_allowed(self, model):
        return True

    def is_region_blocked_in_child(self, child_name:Union[str, Identifier]):
        return self._get_child(child_name).is_region_blocked

class DataModelWalker(ModelWalker):
    _expected_dds = (Model.StatementType.container_, Model.StatementType.list_, Model.StatementType.leaf_, Model.StatementType.leaf_list_)
    _recursive_visited_dds = (Model.StatementType.module_, Model.StatementType.submodule_, Model.StatementType.uses_, Model.StatementType.augment_, Model.StatementType.choice_, Model.StatementType.case_)

class FilteredDataModelWalker(DataModelWalker):
    def _is_allowed(self, model):
        return any(i in model.name.prefix for i in ("nokia", "openconfig"))



# Copyright 2021 Nokia

from typing import Dict, Iterable

from .identifier import Identifier

class ModelPath:
    """Representation of path as a sequence of identifiers.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, parts: Iterable[Identifier]):
        self._path = tuple(parts)

    def is_valid(self, *, only_absolute_path: bool):
        if not self._path:
            return False

        for p in self._path:
            if p.is_valid():
                continue
            if only_absolute_path or p.name != "..":
                return False
        return True

    def repr_path(self):
        return self._path.__repr__()

    def __hash__(self):
        return hash(self._path)

    def __eq__(self, other):
        return isinstance(other, ModelPath) and self._path == other._path



# Copyright 2021 Nokia

import operator

from abc import ABC, abstractmethod

from .errors import *

__all__ = ("Container", "Leaf", "LeafList")

__doc__ = """This module contains wrappers describing the YANG 
structure and metadata obtained from SR OS.

.. Reviewed by PLM 20211201
.. Reviewed by TechComms 20211202
"""

class Schema:
    """Metadata about the YANG schema associated with elements in the data structure.

    .. note:: :py:class:`pysros.wrappers.Schema` metadata is read-only.

    :ivar module: YANG module name from which this node originates

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """
    __slots__ = ("_module",)

    def __init__(self, module):
        self._module = module

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self._module == other._module

    @property
    def module(self):
        """YANG module name from which this node originates.

        .. TODO: Consider removing this property from the documentation as its mentioned as an ivar above - PLM

        .. Reviewed by PLM 20210630
        .. Reviewed by TechComms 20210712
        """
        return self._module

class Wrapper(ABC):
    """Common functionality to support wrappers that describe the YANG structure from
    the SR OS schema.

    .. warning::
       Instance of this class SHOULD NOT be created by user of pysros library.

    .. Reviewed by TechComms 20210712   
    """

    __slots__ = ('_data', '_module')

    def __init__(self, value):
        self.data    = value
        self._module = None

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def __setattr__(self, attr, value):
        if attr == "data" or attr == "schema" or attr in Wrapper.__slots__:
            object.__setattr__(self, attr, value)
        else:
            setattr(self._data, attr, value)

    def __delattr__(self, attr):
        raise make_exception(pysros_err_attr_cannot_be_deleted, obj=self.__class__.__name__, attribute=attr)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self._data == other._data

    def __bool__(self):
        return bool(self._data)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._check_data_type(value)
        self._data = value

    @property
    def schema(self):
        return Schema(self._module) if self._module else None

    @schema.setter
    def schema(self, _):
        raise make_exception(pysros_err_attr_is_read_only, obj=self.__class__.__name__, attribute = 'schema')

    @staticmethod
    @abstractmethod
    def _check_data_type(value):
        pass

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    # helper used in followed loop and deleted
    def _unwrap_first(op):
        return lambda w1, w2: op(w1._data, w2)

    for op in ('__add__', '__sub__', '__mul__', '__truediv__', '__floordiv__'
                , '__mod__', '__pow__', '__lshift__', '__rshift__'
                , '__and__', '__xor__', '__or__'):
        locals()[op] = _unwrap_first(getattr(operator, op))

    # helper used in followed loop and deleted
    def _runwrap_first(op):
        return lambda w1, w2: op(w2, w1._data)

    for rop in ('__radd__', '__rsub__', '__rmul__', '__rtruediv__', '__rfloordiv__'
                , '__rmod__', '__rpow__', '__rlshift__', '__rrshift__'
                , '__rand__', '__rxor__', '__ror__'):
        assert rop.startswith('__r')
        op = '__' + rop[3:]
        locals()[rop] = _runwrap_first(getattr(operator, op))


    # helper used in followed loop and deleted
    def _unwrap_both(op):
        def fn(self, other):
            if self.__class__ is not other.__class__:
                return NotImplemented
            return op(self._data, other._data)
        return fn

    for op in ('__lt__', '__le__', '__gt__', '__ge__'):
        locals()[op] = _unwrap_both(getattr(operator, op))

    del _unwrap_first
    del _runwrap_first
    del _unwrap_both
    del op
    del rop

    @classmethod
    def _with_module(cls, data, module):
        if module and not isinstance(module, str):
            raise make_exception(pysros_err_arg_must_be_string)
        obj = cls(data)
        obj._module = module
        return obj

class Container(Wrapper):
    """YANG container data structure node wrapper.

    A YANG container in the pySROS data structure behaves in the same way as a Python dict,
    where the value of a field can be obtained using the field name.

    :ivar data: dictionary containing the fields data

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """
    __slots__ = ()
    def __init__(self, value = None):
        super().__init__(value if value is not None else {})

    @classmethod
    def _check_data_type(cls, value):
        if not isinstance(value, dict):
            raise make_exception(pysros_err_unsupported_type_for_wrapper, wrapper_name=cls.__name__)

class LeafList(Wrapper):
    """YANG leaf-list data structure node wrapper.

    A YANG leaf-list in the pySROS data structure behaves in the same way as a Python list, where the separate
    values can be obtained using an index.

    :ivar data: list containing the values

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210712
    """
    __slots__ = ()

    def __init__(self, value = None):
        super().__init__(value if value is not None else [])

    @classmethod
    def _check_data_type(cls, values):
        if not isinstance(values, list):
            raise make_exception(pysros_err_unsupported_type_for_wrapper, wrapper_name=cls.__name__)
        for value in values:
            Leaf._check_data_type(value)

class Leaf(Wrapper):
    """YANG leaf data structure node wrapper.

    Depending on the base (root) YANG type of the node, this object wraps a str, int, or bool.

    :ivar data: python object corresponding to the value

    .. Reviewed by PLM 20210630
    .. Reviewed by TechComms 20210713
    """
    __slots__ = ()

    @staticmethod
    def _check_data_type(data):
        if not isinstance(data, (str, int, bool, _Empty)):
            raise make_exception(pysros_err_unsupported_type_for_wrapper, wrapper_name = 'Leaf')


class _Singleton(type):
    _instances = {}
    def __new__(cls, *args, **kwargs):
        res = super(_Singleton, cls).__new__(cls, *args, **kwargs)
        res.__copy__     = lambda self: self
        res.__deepcopy__ = lambda self, memo: self
        res.__reduce__   = lambda self: (self.__class__, ())
        return res

    def __call__(cls, *args, **kwargs):
        assert args == () and kwargs == {}
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class _Empty(metaclass=_Singleton):
    """Representation of whether empty leaf is present.
    
    .. Reviewed by TechComms 20210712
    """
    def __str__(self):
        return "Empty"

    def __repr__(self):
        return "Empty"

Empty = _Empty()
Empty.__doc__="""Define the YANG Empty type."""



# Copyright 2021 Nokia

import copy
import pprint

from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Union
from contextlib import ExitStack

from lxml import etree

from .errors import *
from .identifier import NoModule, Identifier
from .model import Model
from .model_walker import FilteredDataModelWalker, ModelWalker
from .wrappers import *
from .wrappers import _Singleton, Wrapper

_get_tag = lambda x: etree.QName(x).localname

class RequestData:
    """Basic API for holding and handling data.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, root:Model, ns_map:dict):
        self._root = root
        self._data = _ListStorage(root)
        self._ns_map = ns_map

    def process_path(self, path:Union[str, FilteredDataModelWalker], *, strict=False):
        """Create all entries in given path and return setter for last section of the path.

        .. Reviewed by TechComms 20210712
        """
        walker = path if isinstance(path, ModelWalker) else FilteredDataModelWalker.user_path_parse(self._root, path)
        current = _ASetter.create_setter(self._data)
        for elem, keys in zip(walker.path, walker.keys):
            if not isinstance(current, _MoDataSetter):
                raise make_exception(pysros_err_missing_keys, element=current._walker.get_name())
            if elem.data_def_stm not in (Model.StatementType.leaf_, Model.StatementType.leaf_list_):
                if strict and not current.child_mos.is_created(elem.name.name):
                    raise make_exception(pysros_err_no_data_found)
                current = current.child_mos.get_or_create(elem.name.name)
                if keys:
                    if strict and not current.entry_exists_nocheck(keys):
                        raise make_exception(pysros_err_no_data_found)
                    current = current.entry_nocheck(keys)
            elif current.keys.can_contains(elem.name.name):
                current = current.keys.get(elem.name.name)
            else:
                if not current.fields.contains(elem.name.name):
                    if strict:
                        raise make_exception(pysros_err_no_data_found)
                    else:
                        current.fields.set_placeholder(elem.name.name)
                current = current.fields.get(elem.name.name)

        return current

    def set_as_xml(self, value):
        """Populate storage from xml value.

        .. Reviewed by TechComms 20210712
        """
        d = value.xpath("/ncbase:rpc-reply/ncbase:data", namespaces={"ncbase": "urn:ietf:params:xml:ns:netconf:base:1.0"})
        if len(d) != 1:
            raise make_exception(pysros_err_wrong_netconf_response)
        d = d[0]

        root = _ASetter.create_setter(self._data)
        root.set_as_xml(d)

    def to_xml(self):
        """Return storage as xml.

        .. Reviewed by TechComms 20210712
        """
        root = etree.Element("root")
        self._data.to_xml(self._ns_map, root)
        return list(root[0])

    def debug_to_model(self):
        """Print storage as model.

        .. Reviewed by TechComms 20210712
        """
        model = self._data.to_model()
        pprint.pprint(model)

    def debug_dump(self):
        """Dump rquest data in raw form.

        .. Reviewed by TechComms 20210712
        """
        self._data.debug_dump()


class _AStorage(ABC):
    """Abstract representation of data storage.

    .. Reviewed by TechComms 20210712
    """

    @abstractmethod
    def to_xml(self, ns_map, root):
        """Return data in xml format.

        .. Reviewed by TechComms 20210712
        """
        pass

    @abstractmethod
    def to_model(self, *, config_only=False):
        """Return data as user model.

        .. Reviewed by TechComms 20210712
        """
        pass

    @property
    def _walker(self):
        return FilteredDataModelWalker(self._model)

    def _resolve_xml_name(self, model, ns_map):
        if model.name.prefix in ns_map:
            return etree.QName(ns_map[model.name.prefix], model.name.name)
        elif model.name.prefix is NoModule():
            return model.name.name
        else:
            raise make_exception(pysros_err_prefix_does_not_have_ns, prefix=model.name.prefix, name=model.name.name)

def subelement(parent, tag, text=None, attrib=None):
    """Wrapper around etree.Subelement, that raises SrosMgmtError instead of ValueError."""
    try:
        result = etree.SubElement(parent, tag, attrib)
        if text:
            result.text = text
        return result
    except ValueError as err:
        raise SrosMgmtError(err.args) from None

class _MoStorage(_AStorage):
    """Data storage for containers and list entries(list+local keys).

    .. Reviewed by TechComms 20210712
    """

    def __init__(self, model:Model, local_keys):
        self._model = model
        self._local_keys = copy.deepcopy(local_keys)
        self._child = {}
        self._delete = False

    def to_xml(self, ns_map, root):
        root_attr = None if not self._delete else {"operation": "remove"}
        root = subelement(root, self._resolve_xml_name(self._model, ns_map), None, root_attr)
        for k, v in self._local_keys.items():
            if v is FieldValuePlaceholder():
                self._leaf_placeholder_to_xml(v, root, self._walker.get_child(k), ns_map)
            else:
               txt  = self._walker.get_child(k).get_type().to_string(v)
               subelement(root, self._resolve_xml_name(self._walker.get_child(k).current, ns_map), txt)

        for k, v in self._child.items():
            if isinstance(v, (_MoStorage, _ListStorage)):
                v.to_xml(ns_map, root)
            else:
                if v is Delete():
                    subelement(root, self._resolve_xml_name(self._walker.get_child(k).current, ns_map), None, {"operation": "remove"})
                elif v is FieldValuePlaceholder():
                    self._leaf_placeholder_to_xml(v, root, self._walker.get_child(k), ns_map)
                else:
                    self._leaf_to_xml(v, root, self._walker.get_child(k), ns_map)

    def _leaf_to_xml(self, value, root, walker, ns_map):
        if walker.get_dds() == Model.StatementType.leaf_:
            value = (value, )
        for i in value:
            txt = walker.get_type().to_string(i)
            subelement(root, self._resolve_xml_name(walker.current, ns_map), txt)

    def _leaf_placeholder_to_xml(self, value, root, walker, ns_map):
        subelement(root, self._resolve_xml_name(walker.current, ns_map))

    def to_model(self, *, config_only=False):
        if config_only and self._walker.is_state:
            raise make_exception(pysros_err_no_data_found)
        data = {}
        for k, v in self._child.items():
            if config_only and self._walker.get_child(k).is_state:
                continue
            if not isinstance(v, (_MoStorage, _ListStorage)):
                data[k] = self._leaf_to_model(k, v)
            else:
                data[k] = v.to_model(config_only=config_only)
                if not data[k] and not self._walker.get_child(k).has_explicit_presence():
                    del data[k]
        for k, v in self._local_keys.items():
            data[k] = self._leaf_to_model(k, v)
        return Container._with_module(data, self._model.name.prefix)

    def _leaf_to_model(self, name, value):
        assert value is not FieldValuePlaceholder()
        module = self._walker.get_child_name(name).prefix
        if self._walker.get_child_dds(name) == Model.StatementType.leaf_:
            return Leaf._with_module(value, module)
        return LeafList._with_module(value, module)

    def keys_equal(self, keys):
        """Compare if keys are equal. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        return self._local_keys == keys

    def get_keys_flat(self):
        keys = tuple(self._local_keys[key] for key in self._model.local_keys)
        if len(keys) == 1: keys = keys[0]
        return keys

    def debug_dump(self, indent=0):
        line = " "*indent
        if self._delete:
            line += "delete "
        line += self._walker.get_name().name
        if self._local_keys:
            line += " ["
            line += ", ".join(f"{k}={v}" for k, v in self._local_keys.items())
            line += "]"
        line += " {"
        print(line)
        for k, v in self._child.items():
            if isinstance(v, _AStorage):
                v.debug_dump(indent+1)
            else:
                print(f"{' '*(indent+1)}{k} = {v}")
        print((" "*indent)+"}")


class _ListStorage(_AStorage):
    """Storage for all entries for a specific list.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, model:Model):
        if model.data_def_stm == Model.StatementType.container_:
            self.__class__ = _MoStorage
            self.__class__.__init__(self, model, {})
        else:
            self._model = model
            self._entries = {}
            self._delete = False

    def to_xml(self, ns_map, root):
        if self._entries:
            for entry in self._entries.values():
                entry.to_xml(ns_map, root)
        else:
            subelement(root, self._resolve_xml_name(self._model, ns_map))

    def to_model(self, *, config_only=False):
        if config_only and self._walker.is_state:
            raise make_exception(pysros_err_no_data_found)
        d = OrderedDict() if self._model.user_ordered else {}
        for e in self._entries.values():
            d[e.get_keys_flat()] = e.to_model(config_only=config_only)

        return d

    def get_entry(self, keys):
        """Get entry by keys. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        res = self._entries.get(tuple(sorted(keys.items())), None)
        if res is None:
            raise make_exception(pysros_err_entry_does_not_exists)
        return res

    def get_or_create_entry(self, keys):
        """Get entry by keys. Create the entry if it does not exist. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        try:
            return self.get_entry(keys)
        except KeyError:
            return self.add_entry(keys)

    def has_entry(self, keys):
        """Determine if entry with specific keys exists. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        try:
            self.get_entry(keys)
            return True
        except KeyError:
            return False

    def add_entry(self, keys):
        """Add new entry with specific keys. Entry cannot already exist. Keys are expected as dict(name->value).

        .. Reviewed by TechComms 20210712
        """
        assert not self.has_entry(keys)
        res = _MoStorage(self._model, keys)
        self._entries[tuple(sorted(keys.items()))] = res
        return res

    def debug_dump(self, indent=0):
        for i in self._entries.values():
            i.debug_dump(indent+1)

class EntryKeysDictProxy():
    """Proxy for dict keys which unwraps keys and compares by performing Identifier.__eq__"""
    def __init__(self, data):
        if not isinstance(data, dict):
            raise make_exception(pysros_err_invalid_value, data=data)
        self.data = data

    def __getitem__(self, key):
        #need to call Identifier.__eq__
        for k in self.data.keys():
            if key == k:
                return self._unwrap(self.data[k])
        raise KeyError(key)

    def __contains__(self, val):
        #need to call Identifier.__eq__
        for ele in self.data:
            if val == ele:
                return True
        return False

    def __str__(self):
        return str(self.data)

    def _unwrap(self, val):
        return val.data if isinstance(val, Wrapper) else val


class _ASetter(ABC):
    """Data setter representation for storage classes.

    .. Reviewed by TechComms 20210712
    """
    def __init__(self, storage:"_AStorage"):
        self._storage = storage

    @property
    def _walker(self) -> FilteredDataModelWalker:
        return FilteredDataModelWalker(self._storage._model)

    @staticmethod
    def create_setter(storage:"_AStorage"):
        if isinstance(storage, _MoStorage):
            return _MoDataSetter(storage)
        else:
            return _ListSetter(storage)

    @abstractmethod
    def set(self, value):
        """Set data in model format.

        .. Reviewed by TechComms 20210712
        """
        pass

    @abstractmethod
    def set_as_xml(self, value):
        """Set data in xml format.

        .. Reviewed by TechComms 20210712
        """
        pass

    def entry_placeholder(self):
        """Set list keys as a placeholders"""
        raise make_exception(pysros_err_target_should_be_list)


    def to_model(self, *, config_only=False):
        """Return data in model format.

        .. Reviewed by TechComms 20210712
        """
        return self._storage.to_model(config_only=config_only)

    def delete(self):
        self._storage._delete = True

    def _unwrap(self, value):
        return value.data if isinstance(value, Wrapper) else value


    def _as_storage_type(self, val, *, child_name=None):
        walker = self._walker.get_child(child_name) if child_name else self._walker
        if walker.get_dds() == Model.StatementType.leaf_list_:
            return [walker.get_type().as_storage_type(v) for v in val]
        else:
            return walker.get_type().as_storage_type(val)

class _LeafSetter(_ASetter):
    """Interface for managing leafs. Because leafs do not have dedicated storage, the
       class has its own implementation of to_model method.

    .. Reviewed by TechComms 20210713
    """
    def __init__(self, storage:"_ListStorage", leaf_name:str):
        super().__init__(storage)
        self._leaf_name = leaf_name

    def set(self, value):
        value = self._unwrap(value)
        if not self._walker.check_field_value(value):
            raise make_exception(pysros_err_incorrect_leaf_value, value=value, leaf_name=self._leaf_name)
        else:
            self._set_nocheck(self._as_storage_type(value))

    def _set_nocheck(self, value):
        self._storage._child[self._leaf_name] = value

    def set_placeholder(self):
        self._storage._child[self._leaf_name] = FieldValuePlaceholder()

    def set_as_xml(self, value):
        value = value.text or ""
        self.set(self._walker.as_model_type(value))

    def set_or_append_as_xml(self, value):
        value = self._walker.as_model_type(value.text or "")
        if self._walker.get_dds() == Model.StatementType.leaf_list_ and self._leaf_name in self._storage._child:
            value = self._storage._child[self._leaf_name] + value
        self.set(value)

    def to_model(self, *, config_only=False):
        if config_only and self._walker.is_state:
            raise make_exception(pysros_err_no_data_found)
        return self._storage._leaf_to_model(self._leaf_name, self._storage._child[self._leaf_name])

    def delete(self):
        self._storage._child[self._leaf_name] = Delete()

    @property
    def _walker(self):
        return super()._walker.get_child(self._leaf_name)

class _KeySetter(_ASetter):
    """Interface for keys. Most of this class are stub methods to provide
       setter interface to keys."""
    def __init__(self, storage:"_ListStorage", key_name:str):
        super().__init__(storage)
        self._key_name = key_name

    def set(self, value):
        value = self._unwrap(value)
        if not self._walker.check_field_value(value):
            raise make_exception(pysros_err_incorrect_leaf_value, value=value, leaf_name=self._key_name)
        elif self._as_storage_type(value) != self._storage._local_keys[self._key_name]:
            raise make_exception(pysros_err_key_val_mismatch, key_name=self._key_name)

    def set_placeholder(self):
        pass

    def to_model(self, *, config_only=False):
        if config_only and self._walker.is_state:
            raise make_exception(pysros_err_no_data_found)
        return self._storage._leaf_to_model(self._key_name, self._storage._local_keys[self._key_name])

    def delete(self):
        raise make_exception(pysros_err_invalid_operation_on_key)

    def set_as_xml(self, value):
        value = value.text or ""
        self.set(self._walker.as_model_type(value))

    @property
    def _walker(self):
        return super()._walker.get_child(self._key_name)

class _ListSetter(_ASetter):
    """Interface for managing instances of specific list.

    .. Reviewed by TechComms 20210712
    """

    def set(self, value):
        """Set a value in a list.

        Keys may be in format {(key1, key2) : Container({}), (key1, key2) : Container({})}
        or directly in entry {key1: value1, key2: value2, field1 : value}.

        .. Reviewed by PLM 20211018
        """
        is_wrapped = isinstance(value, Container)
        value = self._unwrap(value)
        value = copy.copy(value)
        unwrapper = EntryKeysDictProxy(value)
        if value and not is_wrapped and self._walker.dict_keys(value):
            for k, v in value.items():
                self._handle_entry_keys_namespaces(v)
                self.entry(self._tuple_to_dict(k)).set(v)
        elif self._walker.entry_keys(unwrapper):
            self._handle_entry_keys_namespaces(value)
            self.entry(value).set(value)
        else:
            raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=value)

    def _tuple_to_dict(self, t):
        return {k:v for k, v in zip(self._walker.get_local_key_names(), (t if isinstance(t, tuple) else (t, )))}

    def _extract_keys(self, entry):
        return {k:self._as_storage_type(v, child_name=k) for k, v in entry.items() if k in self._walker.get_local_key_names()}

    def _convert_keys_to_model(self, entry):
        try:
            return {k:self._walker.as_child_model_type(k, v) for k, v in entry.items()}
        except:
            raise make_exception(pysros_err_invalid_key_in_path) from None

    def _handle_entry_keys_namespaces(self, entry):
        """Strip namespace prefixes from entry key names.
        Also raise an error if there are two identical entry keys, one with the namespace prefix and other one without it.
        """
        local_keys = [Identifier(self._walker.get_name().prefix, k) for k in self._walker.get_local_key_names()]
        for k in local_keys:
            if k.model_string in entry:
                if k.name in entry:
                    raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=entry[k.model_string])
                entry[k.name] = entry.pop(k.model_string)

    def _check_and_unwrap_keys(self, entry):
        for k in self._walker.get_local_key_names():
            if k not in entry:
                raise make_exception(pysros_err_malformed_keys, full_path=self._walker, value=value)
            entry[k] = self._unwrap(entry[k])
            if not self._walker.check_child_field_value(k, entry[k]):
                raise make_exception(pysros_err_incorrect_leaf_value, value=entry[k], leaf_name=k)
            entry[k] = self._walker.get_child_type(k).as_storage_type(entry[k])

    def entry(self, value):
        """Receive entry with specified keys in value. Keys are expected as dict(name->value). Additional
        fields may be present (no verification for extra fields).

        .. Reviewed by TechComms 20210712
        """
        self._handle_entry_keys_namespaces(value)
        self._check_and_unwrap_keys(value)
        return _MoDataSetter(self._storage.get_or_create_entry(self._extract_keys(value)))

    def entry_nocheck(self, value):
        """Receive entry with specified keys in value without checking for the correct type. Keys are expected
        as dict(name->value). Additional fields may be present (no verification for extra fields).

        .. Reviewed by PLM 20211018
        """
        return _MoDataSetter(self._storage.get_or_create_entry(self._convert_keys_to_model(self._extract_keys(value))))

    def entry_exists_nocheck(self, value):
        """Receive entry with specified keys in value without checking for the correct type. Keys are expected
        as dict(name->value). Additional fields may be present (no verification for extra fields)
        Returns true if entry exists, otherwise false.
        """
        return self._storage.has_entry(self._convert_keys_to_model(self._extract_keys(value)))

    def entry_xml(self, value):
        keys = {}
        for e in value:
            if _get_tag(e) in self._walker.get_local_key_names():
                keys[_get_tag(e)] = e.text or ""
        if set(keys.keys()) != set(self._walker.get_local_key_names()):
            raise make_exception(pysros_err_schema_box_keys_mismatch, schema_keys=self._walker.get_local_key_names(), box_keys=list(keys.keys()))
        return self.entry_nocheck(keys)

    def set_as_xml(self, value):
        self.entry_xml(value).set_as_xml(value)

    def delete(self):
        raise make_exception(pysros_err_invalid_path_operation_missing_keys)

    def entry_placeholder(self):
        keys = {key:FieldValuePlaceholder() for key in self._walker.get_local_key_names()}
        _MoDataSetter(self._storage.get_or_create_entry(keys)).set({})

class _MoDataSetter(_ASetter):
    """Interface managing specific list entry or container.

    .. Reviewed by TechComms 20210712
    """
    class _AChild:
        def __init__(self, setter:"_MoDataSetter"):
            self._setter = setter

        @property
        def _walker(self) -> FilteredDataModelWalker:
            return FilteredDataModelWalker(self._setter._storage._model)

    class _Keys(_AChild):
        """Interface for retrieving and setting keys.

        .. Reviewed by TechComms 20210712
        """
        def set(self, name, value):
            name = Identifier.from_model_string(name).name
            self.get(name).set(value)

        def set_placeholder(self, name):
            self.get(name).set_placeholder()

        def get(self, name):
            if not self.can_contains(name):
                raise make_exception(pysros_err_unknown_child, child_name=name, path=self._walker._get_path())
            return _KeySetter(self._setter._storage, name)

        def contains(self, name):
            return self.can_contains(name)

        def can_contains(self, name):
            return self._walker.has_local_key_named(name)

    class _Fields(_AChild):
        """Interface for retrieving and setting fields.

        .. Reviewed by TechComms 20210712
        """

        def set(self, name, value):
            name = Identifier.from_model_string(name).name
            self.get(name).set(value)

        def set_placeholder(self, name):
            self.get(name).set_placeholder()

        def get(self, name):
            if not self.can_contains(name):
                raise make_exception(pysros_err_unknown_child, child_name=name, path=self._walker._get_path())
            return _LeafSetter(self._setter._storage, name)

        def contains(self, name):
            return name in self._setter._storage._child

        def can_contains(self, name):
            return self._walker.has_field_named(name)


    class _ChildMos(_AChild):
        """Interface for retrieving and setting children.

        .. Reviewed by TechComms 20210712
        """
        def set(self, name, value):
            self.get_or_create(Identifier.from_model_string(name).name).set(value)

        def get_or_create(self, name):
            if self._walker.get_child_dds(name) in (Model.StatementType.container_, Model.StatementType.list_):
                if not self.is_created(name):
                    self._setter._storage._child[name] = _ListStorage(self._setter._walker.get_child(name).current)
                return _ASetter.create_setter(self._setter._storage._child[name])
            else:
                raise KeyError(name)

        def is_created(self, name):
            return self._walker.get_child_dds(name) in (Model.StatementType.container_, Model.StatementType.list_) and name in self._setter._storage._child


    @property
    def keys(self):
        return self._Keys(self)

    @property
    def fields(self):
        return self._Fields(self)

    @property
    def child_mos(self):
        return self._ChildMos(self)

    def set(self, value):
        value = self._unwrap(value)
        if not isinstance(value, dict):
            raise make_exception(pysros_err_invalid_value, data=value)
        children_to_set = set()
        for k, v in value.items():
            if k in self._walker.get_local_key_names():
                self.keys.set(k, v)
            elif self._walker.get_child_dds(k) in (Model.StatementType.leaf_, Model.StatementType.leaf_list_):
                self.fields.set(k, v)
            elif self._walker.get_child_dds(k) in (Model.StatementType.container_, Model.StatementType.list_):
                self.child_mos.set(k, v)
            else:
                raise make_exception(pysros_err_unknown_dds, dds=self._walker.get_child_dds(k))
            name = Identifier.from_model_string(k).name
            if name in children_to_set:
                raise make_exception(pysros_err_duplicate_found, duplicate=name)
            children_to_set.add(name)

    def set_as_xml(self, value):
        for e in value:
            if not self._walker.has_child(_get_tag(e)):
                raise make_exception(pysros_err_unknown_child, child_name=_get_tag(e), path=self._walker._get_path())
            if self._walker.is_region_blocked_in_child(_get_tag(e)):
                continue
            else:
                if self._walker.get_child_dds(_get_tag(e)) in (Model.StatementType.leaf_, Model.StatementType.leaf_list_) and _get_tag(e) not in self._walker.get_local_key_names():
                    self.fields.get(_get_tag(e)).set_or_append_as_xml(e)
                elif self._walker.get_child_dds(_get_tag(e)) in (Model.StatementType.list_, Model.StatementType.container_):
                    self.child_mos.get_or_create(_get_tag(e)).set_as_xml(e)


class FieldValuePlaceholder(metaclass=_Singleton):
    pass

class Delete(metaclass=_Singleton):
    pass



# Copyright 2021 Nokia

import base64

from abc import ABC, abstractmethod
from collections import OrderedDict
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Set, Type, Tuple, Union

from lxml.etree import SubElement

from .errors import *
from .identifier import Identifier, NoModule
from .model_path import ModelPath
from .wrappers import Empty


INTEGRAL_LEAF_TYPE = ("int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64")
YANG_PRIMITIVE_TYPES = INTEGRAL_LEAF_TYPE + ("binary", "boolean", "decimal64", "empty", "string", "instance-identifier")

class YangTypeBase(ABC):
    """Abstract interface for all leaf types.

    .. Reviewed by TechComms 20210713
    """
    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def __hash__(self):
        pass

    @abstractmethod
    def __eq__(self, other):
        if other is None:
            return False
        assert isinstance(other, YangTypeBase)
        return self.__class__ is other.__class__

    def to_string(self, val: Any) -> str:
        """Translate value to xml text without any checking."""
        # User input is also checked by check_field_value, paths are not
        if isinstance(val, bool) and isinstance(self, PrimitiveType):
            if self.identifier.name == 'boolean':
                return str(val).lower()
            assert self.identifier.name in INTEGRAL_LEAF_TYPE
            return str(int(val))
        if isinstance(val, (int, str)):
            return str(val)
        if val is Empty:
            return ''
        if isinstance(val, bytes):
            return base64.b64encode(val).decode('utf8')
        raise make_exception(pysros_err_unexpected_value_of_type, val_type=type(val), type=str(self))

    def to_value(self, val: str) -> Any:
        """Transform string to native Python type.
            If the value does not fit the expected type, a TypeError exception is raised. This exception is tested
            in union."""
        raise make_exception(pysros_err_invalid_value_for_type, type=self, value=val)

    @abstractmethod
    def check_field_value(self, value: Any) -> bool:
        """Check whether value is acceptable for a specific yang_type."""
        return False

    def json_name(self) -> str:
        return self.__class__.__name__.lower()

    def as_storage_type(self, obj):
        return obj


def _is_valid_decimal64(value: str) -> bool:
    return isinstance(value, str)

def _is_valid_base64(value: Any) -> bool:
    return isinstance(value, (bytes, str))

class PrimitiveType(YangTypeBase):
    """Primitive built-in types representation, such as int8, str, and binary. All types are in YANG_PRIMITIVE_TYPES.

    .. Reviewed by TechComms 20210713
    """
    def __init__(self, name):
        assert name in YANG_PRIMITIVE_TYPES
        self.identifier = Identifier.builtin(name)

    def __str__(self):
        return str(self.identifier)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.identifier.name!r})"

    def __eq__(self, other):
        return YangTypeBase.__eq__(self, other) and self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def to_value(self, val: str) -> Any:
        t = self.identifier.name
        if t in INTEGRAL_LEAF_TYPE:
            return int(val)
        if t == "boolean":
            if val.lower() in ("true", "false"):
                return val.lower() == "true"
        if t == "empty":
            return Empty
        if t in ("string", "binary"):
            return val
        if t == "decimal64":
            return val
        return super().to_value(val)

    _check_field_value = {
        "string":   lambda val: type(val) == str,
        "empty":    lambda val:  val is Empty,
        "boolean":  lambda val:  isinstance(val, (bool, int)),
        "decimal64": _is_valid_decimal64,
        "binary": _is_valid_base64
    }

    def check_field_value(self, value: Any) -> bool:
        name = self.identifier.name
        cmp = self._check_field_value.get(name, None)
        if cmp is not None:
            return cmp(value)

        if name in INTEGRAL_LEAF_TYPE:
            return isinstance(value, (bool, int))
        return False

    def json_name(self) -> str:
        return self.identifier.name

    def as_storage_type(self, obj):
        if self.identifier.name == "boolean" and type(obj) == int:
            return bool(obj)
        elif self.identifier.name in INTEGRAL_LEAF_TYPE and type(obj) == bool:
            return int(obj)
        return super().as_storage_type(obj)

class UnresolvedIdentifier(YangTypeBase):
    """Identifier that has not been resolved yet."""
    def __init__(self, identifier: Identifier):
        self.identifier = identifier
        assert not self.identifier.is_builtin()

    def __str__(self):
        return str(self.identifier)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.identifier!r})"

    def __eq__(self, other):
        return YangTypeBase.__eq__(self, other) and self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def to_string(self, _val: Any) -> str:
        raise make_exception(pysros_err_unresolved_type, type=self)

    def to_value(self, _val: str) -> Any:
        raise make_exception(pysros_err_unresolved_type, type=self)

    def check_field_value(self, value: Any) -> bool:
        return False


class YangUnion(YangTypeBase):
    def __init__(self, types: List["YangType"] = []):
        super().__init__()

        self._types: List["YangType"] = []
        for t in types:
            self.append(t)

    def __iter__(self):
        return self._types.__iter__()

    def __getitem__(self, *args):
        return self._types.__getitem__(*args)

    def append(self, t: "YangType"):
        if t not in self._types:
            self._types.append(t)

    def __str__(self):
        return f"union[{','.join(map(str, self._types))}]"

    def __repr__(self):
        return f"{self.__class__.__name__}(({', '.join(map(repr, self._types))}))"

    def __hash__(self):
        return hash(tuple(self._types))

    def __eq__(self, other):
        return (
            YangTypeBase.__eq__(self, other)
            and self._types == other._types
        )

    def to_string(self, val: Any) -> str:
        if isinstance(val, str):
            return val
        return super().to_string(val)

    def to_value(self, val: str) -> Any:
        return val

    def check_field_value(self, value: Any) -> bool:
        return isinstance(value, str)

    def json_name(self) -> str:
        return "union"

class Enumeration(OrderedDict, YangTypeBase):
    def __str__(self):
        return f"enumeration[{'|'.join(map(str, self.keys()))}]"

    def __hash__(self):
        return hash(tuple(self))

    def add_enum(self, name: str):
        val = 1+max(self.values()) if self else 0
        self[name] = val

    def set_last_enum_value(self, val: int):
        assert self
        last_used_key = next(reversed(self))
        self[last_used_key] = val

    def to_value(self, val: str) -> Any:
        return str(val)

    def check_field_value(self, value: Any) -> bool:
        return isinstance(value, str)


class Bits(set, YangTypeBase):
    def __str__(self):
        return f"bits[{' '.join(sorted(self))}]"

    def __repr__(self):
        return f"Bits(({', '.join(map(repr, sorted(self)))}))"

    def __hash__(self):
        return hash(frozenset(self))

    def __eq__(self, other):
        return YangTypeBase.__eq__(self, other) and frozenset(self) == frozenset(other)

    def is_valid_value(self, val: Any) -> bool:
        return isinstance(val, str)

    def to_value(self, val: str) -> Any:
        if self.is_valid_value(val):
            return val
        return super().to_value(val)

    def check_field_value(self, value: Any) -> bool:
        return self.is_valid_value(value)

class LeafRef(YangTypeBase):
    def __init__(self, path = None):
        if path is None or isinstance(path, ModelPath):
            self._path: Optional[ModelPath] = path
            return
        assert isinstance(path, tuple) and all(isinstance(p, Identifier) for p in path)
        self._path = ModelPath(path)


    def set_path(self, path: ModelPath):
        assert isinstance(path, ModelPath)
        self._path = path

    def __eq__(self, other):
        return YangTypeBase.__eq__(self, other) and self._path == other._path

    def __ne__(self, other):
        return not(self == other)

    def __str__(self):
        return f"leafref({self._path!s})"

    def __repr__(self):
        return f"LeafRef({self._path.repr_path()})"

    def __hash__(self):
        return hash(tuple(self.__class__.__name__, self._path))

    def to_string(self, _val: Any) -> str:
        raise make_exception(pysros_err_unresolved_leafref, type=str(self))

    def to_value(self, _val: str) -> Any:
        raise make_exception(pysros_err_unresolved_leafref, type=str(self))

    def check_field_value(self, value: Any) -> bool:
        return False

    @property
    def path(self):
        return self._path

class IdentityRef(YangTypeBase):
    def __init__(self, bases = (), values: Optional[Set[Identifier]] = None):
        self.bases:  List[Identifier] = list(bases)
        self.values: Set[Identifier] = values or set()

    def add_base(self, base: Identifier):
        assert isinstance(base, Identifier)
        self.bases.append(base)

    def set_values(self, derived: Mapping[Identifier, Set[Identifier]]):
        assert not self.values
        self.values = set()
        for b in self.bases:
            if b not in derived:
                continue
            self.values |= derived[b]

    def __eq__(self, other):
        return YangTypeBase.__eq__(self, other) and self.bases == other.bases and self.values == other.values

    def __str__(self):
        return f"identityref[{', '.join(b.__str__() for b in self.bases)}]"

    def __repr__(self):
        return f"IdentityRef({self.bases!r}, set(({', '.join(sorted(map(repr, self.values)))})))"

    def __hash__(self):
        return hash((self.__class__.__name__, *self.bases, frozenset(self.values)))

    def to_value(self, _val: str) -> Any:
        return _val

    def check_field_value(self, value: Any) -> bool:
        return isinstance(value, str)


YangType = Union[
    PrimitiveType,
    UnresolvedIdentifier,
    Enumeration,
    Bits,
    YangUnion,
    LeafRef,
    IdentityRef]


def resolve_typedefs_shallow(t: YangType, typedefs: Dict[Identifier, YangType]):
    assert isinstance(t, YangTypeBase)
    while type(t) is UnresolvedIdentifier and t.identifier in typedefs:
        t = typedefs[t.identifier]

    assert not isinstance(t, UnresolvedIdentifier), f"Cannot resolve type {t}"
    assert isinstance(t, YangTypeBase)
    return t

def resolve_typedefs_deep(t: YangType, typedefs: Dict[Identifier, YangType]):
    t = resolve_typedefs_shallow(t, typedefs)
    if not isinstance(t, YangUnion):
        return t

    result = YangUnion()
    for sub in t:
        sub = resolve_typedefs_deep(sub, typedefs)
        if isinstance(sub, YangUnion):
            for s in sub:
                result.append(s)
        else:
            result.append(sub)
    return result

_KNOWN_TYPES = {
        "leafref":     LeafRef,
        "bits":        Bits,
        "enumeration": Enumeration,
        "union":       YangUnion,
        "identityref": IdentityRef,
    }

def should_be_buildin(name):
    return name in _KNOWN_TYPES or name in YANG_PRIMITIVE_TYPES

def type_from_name(identifier: Identifier):
    if identifier.is_builtin():
        t = _KNOWN_TYPES.get(identifier.name)
        if t is not None:
            return t()
        return PrimitiveType(identifier.name)
    return UnresolvedIdentifier(identifier)



# Copyright 2021 Nokia

import copy
import locale
import sys

from enum import Enum, IntFlag, IntEnum, auto
from typing import Any, Dict, List, Optional, Tuple

from lxml import etree

from .errors import *
from .identifier import Identifier, NoModule
from .yang_type import Enumeration, YangType, YangTypeBase
from .yang_type import YangType

class AModel:
    class StatementType(Enum):
        leaf_         = 0
        list_         = 1
        container_    = 2
        leaf_list_    = 3
        choice_       = 4
        case_         = 5
        augment_      = 6
        uses_         = 7
        typedef_      = 8
        module_       = 9
        submodule_    = 10
        grouping_     = 11
        import_       = 12
        identity_     = 13
        action_       = 14
        anydata_      = 15
        anyxml_       = 16
        notification_ = 17
        rpc_          = 18
        input_        = 19
        output_       = 20
        deviation_    = 21
        deviate_      = 22

    __slots__ = ()

    def recursive_walk(self, cb):
        cont = cb(self)
        if cont == False:
            return
        for i in self.children:
            i.recursive_walk(cb)

    def __str__(self):
        return f"""Model("{str(self.name)}")"""

    def __repr__(self):
       return f"""Model("{str(self.name)}")"""

    def __deepcopy__(self, memo):
        raise make_exception(pysros_err_use_deepcopy)

    def debug_print(self, prefix="", last = False):
        class Colors:
            RED    = '\033[1;31;48m'
            GREEN  = '\033[1;32;48m'
            DGREEN = '\033[2;32;48m'
            BLUE   = '\033[1;34;48m'
            CYAN   = '\033[1;36;48m'
            PURPLE = '\033[1;35;48m'
            YELLOW = '\033[1;33;48m'
            RESET  = '\033[1;37;0m'

        utf_supported = locale.getlocale()[1] == "UTF-8"
        class AsciiArt:
            CHILD =         "+-- " if not utf_supported else "\u251c\u2500\u2500 "
            LAST_CHILD =    "+-- " if not utf_supported else "\u2514\u2500\u2500 "
            VERTICAL_LINE = "   |" if not utf_supported else "   \u2502"

        data_def_stm = self.data_def_stm.name[:-1]

        if sys.stdout.isatty():
            colorize = lambda data, color : color + str(data) + Colors.RESET
        else:
            colorize = lambda data, color : str(data)

        if not prefix:
            field_prefix = ""
        elif last:
            field_prefix = AsciiArt.LAST_CHILD
        else:
            field_prefix = AsciiArt.CHILD
        if data_def_stm == "container":
            t = colorize(data_def_stm, Colors.YELLOW)
        elif data_def_stm == "list":
            t = colorize(data_def_stm, Colors.GREEN)
        else:
            t = colorize(data_def_stm, Colors.CYAN)
        t_sufix = []
        if data_def_stm in ("leaf", "typedef", ):
            t_sufix.append(colorize(self.yang_type, Colors.BLUE))
        elif data_def_stm == "identity":
            bases = [str(b) for b in self.identity_bases]
            if bases:
                t_sufix.append(colorize(f"base=[{','.join(bases)}]", Colors.BLUE))
        elif data_def_stm == "list":
            t_sufix.append(colorize(",".join(self.local_keys), Colors.DGREEN))
            if self.user_ordered:
                t_sufix.append(colorize("user ordered", Colors.PURPLE))
        elif data_def_stm == "container":
            if self.presence_container:
                t_sufix.append(colorize("presence", Colors.PURPLE))
        elif data_def_stm == "typedef":
            t_sufix.append(colorize(f"->{self.yang_type}", Colors.YELLOW))
        elif data_def_stm == "augment":
            t_sufix.append(colorize(self.target_path, Colors.PURPLE))

        print(f"""{prefix[:-1]}{field_prefix}{self.debug_flags()} {self.name} [{" ".join((t,*t_sufix))}]""")

        remain = self.children_size - 1
        for i in self.children:
            new_prefix = prefix + ("    " if not remain else AsciiArt.VERTICAL_LINE)
            if not prefix:
                new_prefix = new_prefix[3:]
            i.debug_print(new_prefix, not remain)
            remain -= 1

    def debug_flags(self) -> str:
        """Return text flags:
            rw  for configuration data
            ro  for non-configuration data
            -x  for rpcs and actions
            -n  for notifications
            mp   for schema mount points

        .. Reviewed by TechComms 20210713
        """
        data_def_stm = self.data_def_stm.name[:-1]
        if data_def_stm in ("leaf", "list", "leaf_list", "container", "choice", "case"):
            return {
                True:  'rw',
                False: 'ro',
                None:  '??'
            }[self.config]
        if data_def_stm in ("rpc", "action"):
            return "-x"
        if data_def_stm == "notification":
            return "-n"
        return "mp"

    def test_print(self, prefix="", _last = False):
        data_def_stm = self.data_def_stm.name[:-1]
        if not prefix:
            field_prefix = ""
        else:
            field_prefix = "+-- "
        t_sufix = [data_def_stm]
        if data_def_stm in ("leaf", "typedef"):
            t_sufix.append(str(self.yang_type))
        elif data_def_stm == "list":
            t_sufix.append(",".join(self.local_keys))
            if self.user_ordered:
                t_sufix.append("user ordered")
        elif data_def_stm == "container":
            if self.presence_container:
                t_sufix.append("presence")
        elif data_def_stm == "typedef":
            t_sufix.append(f"->{self.yang_type}")
        elif data_def_stm == "augment":
            t_sufix.append(self.target_path)

        res = f"""{prefix[:-1]}{field_prefix}{self.name} [{" ".join(t_sufix)}]\n"""

        remain = self.children_size - 1
        for i in sorted(self.children, key=lambda o: (str(o.name.prefix), o.name.name)):
            new_prefix = prefix + ("    " if not remain else "   |")
            if not prefix:
                new_prefix = new_prefix[3:]
            res += i.test_print(new_prefix, not remain)
            remain -= 1
        return res

class BuildingModel(AModel):
    __slots__ = (
        "name",
        "children",
        "yang_type",
        "presence_container",
        "user_ordered",
        "local_keys",
        "data_def_stm",
        "target_path",
        "identity_bases",
        "_parent",
        "config",
        "blueprint",
    )

    def __init__(self, name: Identifier, data_def_stm: AModel.StatementType, parent):
        self.name = Identifier.builtin(name) if type(name) == str else name
        self.children: List[Model] = []
        self.yang_type: Optional[YangType] = None
        self.presence_container = False
        self.user_ordered = False
        self.local_keys: List[str] = []
        self.data_def_stm = data_def_stm
        self.target_path = None
        self.parent = parent
        self.identity_bases = None if data_def_stm != AModel.StatementType.identity_ else []
        self.config: bool = True
        self.blueprint = []

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, new_parent):
        self._parent = new_parent
        if new_parent:
            new_parent.children.append(self)

    @property
    def has_children(self):
        return bool(self.children)

    @property
    def children_size(self):
        return len(self.children)

    @property
    def prefix(self):
        return self.name.prefix

    @prefix.setter
    def prefix(self, v):
        self.name._prefix = v

    def sanity_parent_child_test(self):
        for i in self.children:
            assert self is i.parent
            i.sanity_parent_child_test()

    def delete_from_parent(self, *, quiet=True):
        idx = self.parent.children.index(self)
        if idx < 0:
            if quiet:
                return
            else:
                raise make_exception(pysros_err_cannot_remove_node, node=self.name)
        del self.parent.children[idx]
        self.parent = None

    def deepcopy(self, parent):
        cls = self.__class__
        result = cls.__new__(cls)
        for i in self.__slots__:
            if i in ("_parent", "children"):
                continue
            setattr(result, i, copy.deepcopy(getattr(self, i)))
        result.parent = parent
        result.children = []
        for child in self.children:
            child.deepcopy(result)
        return result

class StorageModel(AModel):
    __slots__ = (
        "_storage",
        "_index",
    )

    class DataBitmask(IntFlag):
        data_def_stm        = 0xff
        presence_container  = 1 << 8
        user_ordered        = 1 << 9
        config              = 1 << 10

    class DataMembers(IntEnum):
        name            = 0
        children        = 1
        yang_type       = 2
        local_keys      = 3
        target_path     = 4
        identity_bases  = 5
        parent          = 6
        bitmask         = 7

class Model(StorageModel):
    """Class to represent a single YANG entry and hold additional information, such as type, configuration state, children, and data definition statement.

    .. Reviewed by TechComms 20210713
    """
    __slots__ = ("_data", "children", "parent", "name", "_bitmask", "data_def_stm",)

    def __init__(self, storage, index, parent):
        assert storage
        assert isinstance(storage, list)
        assert 0 <= index < len(storage)
        self._storage = storage
        self._index = index
        self._data = self._storage[self._index]
        self.parent = parent

    def __getattr__(self, name):
        getter = {
            "name": Model._name_getter,
            "children": Model._children_getter,
            "_bitmask": Model._bitmask_getter,
            "data_def_stm": Model._data_def_stm_detter,
        }.get(name)
        if not getter:
            raise AttributeError(name)
        res = getter(self)
        setattr(self, name, res)
        return res

    def _name_getter(self):
        name = self._data[Model.DataMembers.name]
        if name[0] is None:
            return Identifier.builtin(name[1])
        return Identifier(*name)

    def _children_getter(self):
        return [Model(self._storage, index, self) for index in self._data[Model.DataMembers.children]]

    def _bitmask_getter(self):
        return self._data[Model.DataMembers.bitmask]

    def _data_def_stm_detter(self):
        return Model.StatementType(int(self._bitmask & Model.DataBitmask.data_def_stm))

    @property
    def prefix(self):
        return self._data[Model.DataMembers.name][0]

    @property
    def has_children(self):
        return bool(self.children_data)

    @property
    def children_size(self):
        return len(self.children_data)

    @property
    def children_data(self):
        return self._data[Model.DataMembers.children]

    @property
    def yang_type(self):
        return self._data[Model.DataMembers.yang_type]

    @property
    def presence_container(self):
        return bool(self._bitmask & Model.DataBitmask.presence_container)

    @property
    def user_ordered(self):
        return bool(self._bitmask & Model.DataBitmask.user_ordered)

    @property
    def is_region_blocked(self):
        return self.prefix in ("nokia-bof-state", "nokia-li-state", "nokia-debug-state", "nokia-li-conf", "nokia-bof-conf", "nokia-debug-conf")

    @property
    def local_keys(self):
        return self._data[Model.DataMembers.local_keys]

    @property
    def target_path(self):
        return self._data[Model.DataMembers.target_path]

    @property
    def identity_bases(self):
        return self._data[Model.DataMembers.identity_bases]

    @property
    def config(self):
        return bool(self._bitmask & Model.DataBitmask.config)

    def sanity_parent_child_test(self):
        for i in self.children:
            assert self._data is i.parent._data
            i.sanity_parent_child_test()

    def __str__(self):
        return f"""Model("{str(self.name)}")"""

    def __repr__(self):
        return f"""Model("{str(self.name)}")"""

    def __deepcopy__(self, memo):
        raise make_exception(pysros_err_use_deepcopy)

    def __eq__(self, other):
        return self._data is other._data

class StorageConstructionModel(StorageModel):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        storage, data = new_Model_data(*args, **kwargs)
        self._storage = storage
        self._index = len(storage)
        self._storage.append(data)
        self.parent = self.parent

    @classmethod
    def new_from_data(cls, storage, index):
        assert storage
        assert isinstance(storage, list)
        assert 0 <= index < len(storage)
        res = cls.__new__(cls)
        res._storage = storage
        res._index = index
        return res

    @property
    def name(self):
        name = self._data[self.DataMembers.name]
        if name[0] is None:
            return Identifier.builtin(name[1])
        return Identifier(*name)

    @property
    def prefix(self):
        return self._data[self.DataMembers.name][0]

    @prefix.setter
    def prefix(self, value):
        self._data[self.DataMembers.name][0] = value

    @property
    def children(self):
        return (self.new_from_data(self._storage, index) for index in self._data[self.DataMembers.children])

    @property
    def has_children(self):
        return bool(self.children_data)

    @property
    def children_size(self):
        return len(self.children_data)

    @property
    def children_data(self):
        return self._data[self.DataMembers.children]

    @children_data.setter
    def children_data(self, data):
        self._data[self.DataMembers.children] = data

    @property
    def yang_type(self):
        return self._data[self.DataMembers.yang_type]

    @yang_type.setter
    def yang_type(self, value):
        self._data[self.DataMembers.yang_type] = value

    @property
    def presence_container(self):
        return bool(self._bitmask & self.DataBitmask.presence_container)

    @presence_container.setter
    def presence_container(self, value):
        if value:
            self._bitmask |= int(self.DataBitmask.presence_container)
        else:
            self._bitmask &= int(~self.DataBitmask.presence_container)

    @property
    def user_ordered(self):
        return bool(self._bitmask & self.DataBitmask.user_ordered)

    @user_ordered.setter
    def user_ordered(self, value):
        if value:
            self._bitmask |= int(self.DataBitmask.user_ordered)
        else:
            self._bitmask &= int(~self.DataBitmask.user_ordered)

    @property
    def local_keys(self):
        return self._data[self.DataMembers.local_keys]

    @local_keys.setter
    def local_keys(self, value):
        self._data[self.DataMembers.local_keys] = value

    @property
    def data_def_stm(self):
        return self.StatementType(int(self._bitmask & self.DataBitmask.data_def_stm))

    @property
    def target_path(self):
        return self._data[self.DataMembers.target_path]

    @target_path.setter
    def target_path(self, value):
        self._data[self.DataMembers.target_path] = value

    @property
    def identity_bases(self):
        return self._data[self.DataMembers.identity_bases]

    @identity_bases.setter
    def identity_bases(self, value):
        self._data[self.DataMembers.identity_bases] = value

    @property
    def parent(self):
        parent = self._data[self.DataMembers.parent]
        if parent is None:
            return parent
        return self.new_from_data(self._storage, parent)

    @parent.setter
    def parent(self, new_parent):
        self._data[self.DataMembers.parent] = new_parent if new_parent is None else new_parent._index
        if new_parent:
            assert new_parent._storage is self._storage
            new_parent.children_data.append(self._index)

    @property
    def config(self):
        return bool(self._bitmask & self.DataBitmask.config)

    @config.setter
    def config(self, value):
        if value:
            self._bitmask |= int(self.DataBitmask.config)
        else:
            self._bitmask &= int(~self.DataBitmask.config)

    @property
    def _bitmask(self):
        return self._data[self.DataMembers.bitmask]

    @_bitmask.setter
    def _bitmask(self, value):
        self._data[self.DataMembers.bitmask] = value

    @property
    def _data(self):
        return self._storage[self._index]

    def sanity_parent_child_test(self):
        for i in self.children:
            assert self._data is i.parent._data
            i.sanity_parent_child_test()

    def __str__(self):
        return f"""Model("{str(self.name)}")"""

    def __repr__(self):
        return f"""Model("{str(self.name)}")"""

    def deepcopy(self, parent):
        cls = self.__class__
        result = cls.__new__(cls)
        result._storage = parent._storage
        result._index = len(result._storage)
        result._storage.append(copy.deepcopy(self._data))
        result.children_data = []
        result.parent = parent
        for child in self.children:
            child.deepcopy(result)
        return result

    def __deepcopy__(self, memo):
        raise make_exception(pysros_err_use_deepcopy)

    def __eq__(self, other):
        return self._data is other._data

def new_Model_data(name: Identifier, data_def_stm: Model.StatementType, parent):
    return (parent._storage if parent is not None else []), [
        [None, name] if type(name) == str else [name.prefix, name.name],        # 0 = name
        [],                                                                     # 1 = children
        None,                                                                   # 2 = yang_type
        [],                                                                     # 3 = local_keys
        None,                                                                   # 4 = target_path
        None if data_def_stm != Model.StatementType.identity_ else [],          # 5 = identity_bases
        parent if parent is None else parent._index,                            # 6 = parent
        data_def_stm.value | Model.DataBitmask.config,                          # 7 = bitmask
    ]
